//////////////////////////////////////////////////////////////////////////// 
// Image_Processing_toolbox_2
//
// Copyright (c) by S. Parua Biswas, CDAC, Kolkata.  2013
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
////////////////////////////////////////////////////////////////////////////

function y = fwht(x,N,ordering)
    // FWHT Fast Discrete Walsh-Hadamard Transform
    // Y = FWHT(X) is the discrete Walsh-Hadamard transform of vector X. The transform coefficients are stored in Y. If X is a matrix, the function operates on each column.
    
    // Y = FWHT(X,N) is the N-point discrete Walsh-Hadamard transform of the vector X where N must be a power of two. X is padded with zeros if it has less than N points and truncated if it has more. The default value of N is equal to the length of the vector X if it is a power of two or the next power of two greater than the signal vector length. The function errors out if N is not equal to a power of two.
    
    // Y = FWHT(X,N,ORDERING) or FWT(X, [],ORDERING) specifies the order of the Walsh-Hadamard transform coefficients. OEDERING can be 'sequency', 'hadamard' or 'dyadic'. Default ORDERING type is 'sequency'. 
    
    [lhs,rhs]=argn(0);
        if ((rhs<1) | (rhs>3)) then
            error('The no of input parameters are not in the range');
        end
    
        if isempty(x) then
            y = [];
            return
        end
        if (rhs<3) then
            ordering = 'sequency'; //default ordering is sequency
            if (rhs<2) then
                N = defaulttransformlength(size(x));
            end
        end
        
        // Input datatype shouldn't be of single precision
        try
            chkinputdatatype(x,N);
        catch
            disp('exception 2');
        end
        
        if isempty(N) then
            N =  defaulttransformlength(size(x));
        end
        
        validatetransformlength(N);
        validateordering(ordering); 
        
        
        // do pre-processing on input signal if necessary
        
[x,tFlag] = preprocessing(x,N,ordering);   

// calculate first stage coefficients and store in x   
       for i = 1:2:N-1
           x(i,:) = x(i,:) + x(i+1,:);
           x(i+1,:) = x(i,:) - 2 * x(i+1,:);
       end
       L = 1;
    //   y = zeros(size(x));
       [rw cl] = size(x);
       for i = 1:rw
           for j = 1:cl
               y(i,j) = 0;
           end
       end

       for nStage = 2:log2(N) // log2(N) = number of stages in the flow diagram
           // calculate coefficients for the ith stage specified by nStage
           M = 2^L;
           J = 0; K = 1;
           while (K < N)
               for j = J+1:2:J+M-1
                   if strcmpi(ordering,'sequency') == 0 then
                       y(K,:) = x(j,:) + x(j+M,:);
                       y(K+1,:) = x(j,:) - x(j+M,:);
                       y(K+2,:) = x(j+1,:) - x(j+1+M,:);
                       y(K+3,:) = x(j+1,:) + x(j+1+M,:);
                       
                   else
                       y(K,:) = x(j,:) + x(j+M,:);
                       y(K+1,:) = x(j,:) - x(j+M,:); 
                       y(K+2,:) = x(j+1,:) + x(j+1+M,:);  
                       y(K+3,:) = x(j+1,:) -  x(j+1+M,:);
                   end // end if
                   K = K + 4;
               end // end for
               
               J = J + 2*M;
           end // end while
           
           // store coefficients in x at the end of each stage
           x = y;
           
           // size of x is changed here
           L = L + 1;
           
       end // end for
       
       // Perform scaling of coefficients
       y = x ./ N;
       if tFlag then
           y = y';
           
       end
      // 
endfunction

//--------------------------------------------------------------------------//

function validatetransformlength(N)
// check if transform length is specified correctly - positive scalar and
// power of two
// sigDim = [rows cols] of input signal

    if ~isempty(find(char(mtlb_num2str(log2(N)))=='.', 1)) | (N < 1) then
        error('Transform length must be a non-negative integer that is a power of 2');
    end
endfunction

//--------------------------------------------------------------------------//

function N = defaulttransformlength(sigDim) 
// this function assigns default transform length as signal length if it is
// a power of two or the next power of two greater than signal length
// sigDim = [rows cols] of input signal

if min(sigDim) == 1 then
    N = max(sigDim);
else
    N = sigDim(1);
end

if ~isempty(strindex(char(mtlb_num2str(log2(N))),'.')) then
  
    p = ceil(log2(N));
    N = 2^p;
    
end

endfunction

//--------------------------------------------------------------------------

function validateordering(ordering) 
// this functions checks if the specified ordering is a string input and one
// of sequency, hadamard or dyadic'
if ((type(ordering)~=10) | ~(strcmpi(ordering,'sequency')==0 | strcmpi(ordering,'hadamard')==0 | strcmpi(ordering,'dyadic')==0)) //~any(strcmpi(ordering,{'sequency','hadamard','dyadic'}))) then

 //   if type(ordering)==10 |  then
 //   end
    error('Output ordering must be sequency, hadamard or dyadic');
end
endfunction
    
//-------------------------------------------
function [x,tFlag] = preprocessing(x,N,ordering)
// this function performs zero-padding, truncation or input bit-reversal if
// necessary. NROWS amd MCOLS specify the output orientation which is kept
// same as that of input.

tFlag = 0;
if size(x,1) == 1 then
    x = x(:); // column vectorizing input sequence
    tFlag = 1;
end

n = size(x,1);

if n < N then

    // zero-pad
    x(n+1:N,:) = 0;
elseif n > N then
    // truncate
    x = x(1:N,:);
end

// Re-arrange input in bit-reversed order if ordering is hadamard

if (strcmpi(ordering,'hadamard') == 0) then
    x = bitrevorder(x);
end


endfunction