//////////////////////////////////////////////////////////////////////////// 
// Image_Processing_toolbox_2
//
// Copyright (c) by S. Parua Biswas, CDAC, Kolkata.  2013
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
////////////////////////////////////////////////////////////////////////////

function [img,H] = iradon(varargin)
    //IRADON - Inverse radon transform
    
    [p,theta,filter,d,interp_var,N] = parse_inputs(varargin(:));
    
    [p,H] = filterProjections(p, filter, d); //
    
    //////////////////////////
    len = size(p,1);
    H = designFilter(filter,len,d);
    
    ////////////////////////////
    // Define the x & y axes for the reconstructed image so that the origin (center) is in the spot which RADON would choose.
    center_var = floor((N + 1)/2);
    xleft = -center_var + 1;
    x = (1:N)-1+xleft;
    x = repmat(x, N, 1);
    
    ytop = center_var - 1;
    y = (N:-1:1).' - N + ytop;
    y = repmat(y, 1, N);
    
    len = size(p,1);
    ctrIdx = ceil(len/2);
    
    imgDiag = 2*ceil(N/sqrt(2))+1; //largest distance through image
    if size(p,1) < imgDiag then
        rz = imgDiag - size(p,1); // how many rows of zeros
        p = [zeros(ceil(rz/2),size(p,2)); p; zeros(floor(rz/2),size(p,2))];
        ctrIdx = ctrIdx+ceil(rz/2);
    end
    
    // Backprojection - vectorized in (x,y), looping over theta
    select interp_var
    case 'nearest neighbor' then img = iradonmex(N, theta, x, y, p, 0);
    case 'linear' then img = iradonmex(N, theta, x, y, p, 1);
    case {'spline','pchip','cubic','v5cubic'} then interp_method = sprintf('*%s',interp_var);
   
            
     // Generate trignometric tables
        costheta = cos(theta);
        sintheta = sin(theta);
    
        // Allocate memory for the image
        img = zeros(N,class(p));
    
        for i=1:length(theta)
            proj_var = p(:,i);
            taxis = (1:size(p,1)) - ctrIdx;
            t = x.*costheta(i) + y.*sintheta(i);
            projContrib = interp1(taxis,proj_var,t(:),interp_method);
            img = img + reshape(projContrib,N,N);        
        end // for
      
  end //select
  img = img*%pi/(2*length(theta));
    
endfunction

//-----------------------------------------------
//--------------Functions------------------------------

function [p,H] = filterProjections(p_in, filter, d)
    p = p_in;
    
    // Design the filter
    len = size(p,1);
    H = designFilter(filter, len, d);
    
    if strcmpi(filter, 'none') then
        return;
    end
    
    p(length(H),1)=0; // Zero pad projections
    
    p = fft(p); // p holds fft of projections
    
    for i = 1:size(p,2)
        p(:,i) = p(:,i).*H; // frequency domain filtering
    end
    
    p = real(ifft(p)); // p is the filtered projections
    p(len+1:end,:) = []; // Truncate the filtered projections
    
endfunction

//------------------------------

function filt = designFilter(filter, len, d)
    // Returns the Fourier Transform of the filter which will be used to filter the projections
    // INPUT ARGS: filter - either the string specifying the filter
               //  len - the length of the projections
               //  d - the fraction of frequencies below the nyquist ----- which we want to pass
   // OUTPUT ARGS: filt - the filter to use on the projections
   
   order = max(64,2^nextpow2(2*len));    
   
   if strcmpi(filter, 'none') then
       filt = ones(1, order);
       return;
   end
   
   // First create a ramp filter - go up to the next highest power of 2.
   
   filt = 2*(0:(order/2))./order;
   w = 2*%pi*(0:size(filt,2)-1)/order; // frequency axis up to Nyquist
   
   select filter
   case 'ram-lak' then ;
   case 'shepp-logan' then filt(2:end) = filt(2:end).* (sin(w(2:end)/(2*d))./(w(2:end)/(2*d))); // not to devide by 0;
case 'cosine' then filt(2:end) = filt(2:end) .* cos(w(2:end)/(2*d));
case 'hamming' then filt(2:end) = filt(2:end) .* (.54 + .46 * cos(w(2:end)/d));
case 'hann' then filt(2:end) = filt(2:end) .* (1+cos(w(2:end)./d)) / 2;
else
    eid = sprintf('Images:%s:invalidFilter',mfilename);
    msg = 'Invalid filter selected.';
    error(msg);
end


filt(w>%pi*d) = 0; // Crop the frequency response
filt = [filt' ; filt(end-1:-1:2)']; // Symmetry of the filter
            
    
endfunction

//------------------------
function [p,theta,filter,d,interp_var,N] = parse_inputs(varargin)
    // Inputs: varargin - Cell array containing all of the actual inputs
    //Outputs: p - Projection data
    //    theta - the angles at which the projections were taken
    // filter - string specifying filter or the actual filter
    // d - a scalar specifying normalized frequency at which to crop the frequency response of the filter
    // interp_var - the type of interpolation to use
    // N - The size of the reconstructed image  
    [lhs,rhs]=argn(0);
    if rhs<2 then
        eid = sprintf('Images:%s:tooFewInputs',mfilename);
        msg = 'Invalid input arguments.';
        error(msg);  
    end
    
    p = varargin(1);
    theta = %pi*varargin(2)/180;
    
    // Default values
    N = 0; // Size of the reconstructed image
    d = 1; // Defaults to no cropping of filters frequency response
    filter = 'ram-lak'; // The ramp filter is the default
    interp_var = 'linear'; // default interpolation is linear
    
    interp_strings = {'nearest neighbor', 'linear', 'spline', 'pchip', 'cubic', 'v5cubic'};
    filter_strings = {'ram-lak','shepp-logan','cosine','hamming','hann','none'};
    string_args = [interp_strings filter_strings];
    
    for i=3:rhs
        arg = varargin(i);
        if ischar(arg) then
            idx = strmatch(lower(arg),string_args);
            if isempty(idx) then
                eid = sprintf('Images:%s:unknownInputString',mfilename);
                msg = sprintf('Unknown input string: %s.', arg);
                error(msg);
            elseif NumOfEle(idx) > 1 then
                eid = sprintf('Images:%s:ambiguousInputString',mfilename);
                msg = sprintf('Ambiguous input string: %s.',arg);
                error(msg);
            elseif NumOfEle(idx) == 1 then
                if idx <=NumOfEle(interp_strings) then
                    interp_var = string_args(idx);
                elseif (idx > NumOfEle(interp_strings)) & (idx <= NumOfEle(string_args)) then
                    filter = string_args(idx);                
                end                
            end
        elseif NumOfEle(arg)==1 then
            if arg <=1 then
                d = arg;
            else 
                N = arg;
            end
        else
            eid = sprintf('Images:%s:InvalidInputParameters',mfilename);  
            msg = 'Invalid input parameters';
            error(msg);         
        end
    end
    
    // If the user didn't specify the size of the reconstruction, so deduce it from the length of projections
    if N==0 then
        N = 2*floor(size(p,1)/(2*sqrt(2)));
    end
    
    if isempty(theta) then
        theta = %pi / size(p,2);        
    end
    
    // If the user passed in delta-theta, build the vector of theta values.
    if NumOfEle(theta)==1 then
        theta = (0:(size(p,2)-1))*theta;
    end
    
    if length(theta) ~= size(p,2) then
        eid = sprintf('Images:%s:thetaNotMatchingProjectionNumber',mfilename);
        msg = 'THETA does not match the number of projections.';
        error(msg); 
    end
    
        
endfunction

//----------------------------------------------------------------------
