//////////////////////////////////////////////////////////////////////////// 
// Image_Processing_toolbox_2
//
// Copyright (c) by S. Parua Biswas, CDAC, Kolkata.  2013
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
////////////////////////////////////////////////////////////////////////////

function [F,ogamma,otheta] = para2fan(varargin)
    args = parseInputs(varargin(:));
    
    P=args.P;
    d=args.d;
    dthetaDeg=args.FanRotationIncrement;
    interp=args.Interpolation;
    
    utils=fanUtils;
    [P,m]=utils.padToOddDim(P);
    
    dploc=args.parallelSensorSpacing;
    ploc=formPlocVector(m,dploc);
    
    fanSpacing=args.FanSensorSpacing;
    [gammaRad,floc]=formGammaVector(ploc,d,fanSpacing,args.FanSensorGeometry);
    gammaDeg=gammaRad*180/%pi;
    
    isParallelCoverageCycle = strcmp(args.ParallelCoverage,'cycle');
    
    n = size(P,2);
    [pthetaDeg,dpthetaDeg] = formPthetaVector(n,isParallelCoverageCycle);
    
    if isempty(dthetaDeg) then
        dthetaDeg = dpthetaDeg;
    end
    
    gammaMax = max(gammaDeg);
    gammaMin = min(gammaDeg);
    
    if strcmp(args.FanCoverage,'minimal') then
        thetaDeg = utils.formMinimalThetaVector(n,dthetaDeg,gammaMin,gammaMax);
    else
        thetaDeg=formCycleThetaVector(dthetaDeg);
    end
    
    numelGamma = numel(gammaDeg);
    
    // interpolate to get desired t sample locations
    // t = d*sin(gammaRad) = distance of projection sample (beam) to iso-center
    
    // interpolate in ploc domain, since this represents actual spacing
    Pint = zeros(numelGamma,n);
    t = d*sin(gammaRad); // t approximates fan beam as parallel beam.
    for i = 1:numel(pthetaDeg)
        Pint(:,i) = interp1(ploc,P(:,i)',t,interp)';
    end
    
    // build padded Pint
    if isParallelCoverageCycle then
        Ppad = Pint;
        pthetapad = pthetaDeg;
    else
        [Ppad, pthetapad] = utils.repPforCycleCoverage(Pint,pthetaDeg);
    end
    gammaRange = gammaMax - gammaMin;
    ptmask = (pthetapad >= 360 - gammaRange);
    ptmask2 = (pthetapad <= gammaRange);
    Ppad = [Ppad(:,ptmask) Ppad Ppad(:,ptmask2)];
    pthetapad = [pthetapad(ptmask)-360 pthetapad pthetapad(ptmask2)+360];
    
    F = shiftAndInterpRotationAngles(gammaDeg,thetaDeg,pthetapad,Ppad,interp);
    
    if isParallelCoverageCycle
        Ppad2 = flipud(Ppad);
        pthetapad2 = pthetapad - 180;
        theta2 = mod(thetaDeg+180,360) - 180;
        
        F2 = shiftAndInterpRotationAngles(gammaDeg,theta2,pthetapad2,Ppad2,interp);
        
        //Avg results
        F = (F+F2)/2;         
    
    end
    
    otheta = thetaDeg;
    if strcmp(args.FanSensorGeometry,'line') then
        ogamma = floc;
    else
        ogamma = gammaDeg;
    end
    ogamma = ogamma';
    
    F = utils.setNaNsToZero(F);    
  
endfunction

//////////////////////////////////////////////

function F = shiftAndInterpRotationAngles(gamma,theta,pthetapad,Ppad,interp)
    
    numelGamma = numel(gamma);
    F = zeros(numelGamma,numel(theta));
    for i=1:numelGamma
        F(i,:) = interp1(pthetapad+gamma(i),Ppad(i,:),theta,interp);
    end
endfunction

//////////////////////////////////////
function [gammaRad,floc] = formGammaVector(ploc,d,fanSpacing,fanSensorGeometry)
    utils = fanUtils;
    floc = [];
    
    plocMax = max(ploc);
    plocMin = min(ploc);
    if strcmp(fanSensorGeometry,'line') then
        floc = utils.formVectorCenteredOnZero(fanSpacing,plocMin,plocMax);
        gammaRad = atan(floc/d);
    else
        gammaRad = utils.formVectorCenteredOnZero(fanSpacing*(%pi)/180,asin(plocMin/d),asin(plocMax/d));
    end
endfunction
    //////////////////////////////////////
    function [ptheta,dpthetaDeg] = formPthetaVector(n,isParallelCoverageCycle)
        
        if isParallelCoverageCycle then
            dpthetaDeg = 360/n;
        else
            dpthetaDeg = 180/n;
        end
        ptheta = dpthetaDeg * (0:n-1);
    endfunction
    
    //////////////////////////////////////////
    function theta = formCycleThetaVector(dthetaDeg)
        
        mfilename = 'formCycleThetaVector';
        if (mod(360,dthetaDeg) ~= 0) then
            eid = sprintf('Images:%s:dthetaNotFactorOf360',mfilename);
            error(eid,'''FanRotationIncrement'' must be a factor of 360.');
        end
        theta = 0:dthetaDeg:(360-dthetaDeg);
    endfunction
    
    ///////////////////////////////////
    function args = parseInputs(varargin)
        
        mfilename = 'parseInputs';
        [lhs,rhs]=argn(0);
        iptchecknargin(2,16,rhs,mfilename);
        
        P = varargin(1);
        d = varargin(2);
        
        args.ParallelSensorSpacing = [];
        args.FanSensorSpacing = [];
        args.FanRotationIncrement = [];
        args.ParallelCoverage = 'halfcycle';
        args.FanSensorGeometry = 'arc';
        args.FanCoverage = 'cycle';
        args.Interpolation = 'linear';
        
        valid_params = {'ParallelSensorSpacing',...
                        'FanSensorSpacing',...
                        'FanRotationIncrement',...
                        'ParallelCoverage',...
                        'FanSensorGeometry',...
                        'FanCoverage',...
                        'Interpolation'};
                        
        num_pre_param_args = 2;
        args = check_fan_params(varargin(3:rhs),args,valid_params,...
                                mfilename,num_pre_param_args);
                                
        if isempty(args.ParallelSensorSpacing) then
            args.ParallelSensorSpacing = 1;
        end
        
        // To compute maxPloc
        m = size(P,1);
        maxPloc = args.ParallelSensorSpacing*(m-1)/2;
        if d < maxPloc then
            msg1 = 'Inconsistency detected among the parameters D, PARALLEL_SENSOR_SPACING,';
            msg2 = 'and FAN_SENSOR_GEOMETRY. Chaeck your input parameters.';
            msg3 = sprintf('Try increasing D to be at least %d.',ceil(maxPloc));
            eid = sprintf('Images:%s:dTooSmall',mfilename);
            error(eid,'%s\n%s\n%s',msg1,msg2,msg3);
        end
        
        if isempty(args.FanSensorSpacing) then
            
            if strcmp(args.FanSensorGeometry,'line') then
                args.FanSensorSpacing = d*sin(args.ParallelSensorSpacing/d);
            else
                args.FanSensorSpacing = asin(args.ParallelSensorSpacing/d)*180/%pi;
            end
        end
        
        args.P = P;
        args.d = d;
        
    endfunction
    
